//******************************************************************************
//
// Microsoft Windows Media
// Copyright (C) Christian Donner @ pod llc. All rights reserved.
//
// FileName: main.cpp
//
// Abstract: File MP3 data into individual directories.
//
//******************************************************************************

#include <tchar.h>
#include <stdio.h>
#include <string.h>
#include <wmsdk.h>
#include <io.h>

//------------------------------------------------------------------------------
// Macros
// ------------------------------------------------------------------------------
#ifndef SAFE_RELEASE

#define SAFE_RELEASE( x )       \
    if( NULL != x )             \
    {                           \
        x->Release( );          \
        x = NULL;               \
    }

#endif // SAFE_RELEASE

#ifndef SAFE_DELETE

#define SAFE_DELETE( x )        \
    if( NULL != x )             \
    {                           \
        delete x;               \
        x = NULL;               \
    }

#endif // SAFE_DELETE

#ifndef SAFE_ARRAYDELETE

#define SAFE_ARRAYDELETE( x )   \
    if( NULL != x )             \
    {                           \
        delete [] x;            \
        x = NULL;               \
    }

#endif // SAFE_ARRAYDELETE


#define TYPE_STANDARD 0
#define TYPE_YEAR 1
#define TYPE_ARTIST 2
#define TYPE_ALBUM 3
#define TYPE_GENRE 4

#define V_VERBOSE 1
#define V_QUICK   0


#define BUFSIZE MAX_PATH
	

//------------------------------------------------------------------------------
// Name: EditorOpenFile()
// Desc: Creates metadata editor, opens file, and gets header info interface.
//------------------------------------------------------------------------------
HRESULT EditorOpenFile( WCHAR* pwszInFile, 
                        IWMMetadataEditor ** ppEditor,
                        IWMHeaderInfo ** ppHeaderInfo, 
                        IWMHeaderInfo3 ** ppHeaderInfo3 )
{
    if( ( NULL == pwszInFile ) ||
        ( NULL == ppEditor ) )
    {
        return( E_INVALIDARG );
    }

    HRESULT hr = S_OK;

    do
    {
        hr = WMCreateEditor( ppEditor );
        if( FAILED( hr ) )
        {
            _tprintf( _T( "Could not create Metadata Editor ( hr=0x%08x ).\n" ), hr );
            break;
        }

        hr = ( *ppEditor )->Open( pwszInFile );
        if( FAILED ( hr ) )
        {
            _tprintf( _T( "Could not open the file %ws ( hr=0x%08x ).\n" ), 
                pwszInFile, hr );
            break;
        }

        if( NULL != ppHeaderInfo )
        {
            hr = ( *ppEditor )->QueryInterface( IID_IWMHeaderInfo, 
                                                (void **)ppHeaderInfo );
            if( FAILED( hr ) )
            {
                _tprintf( _T( "Could not QI for IWMHeaderInfo ( hr=0x%08x ).\n" ), hr );
                break;
            }
        }

        if( NULL != ppHeaderInfo3 )
        {
            hr = ( *ppEditor )->QueryInterface( IID_IWMHeaderInfo3, 
                                                (void **)ppHeaderInfo3 );
            if( FAILED( hr ) )
            {
                _tprintf( _T( "Could not QI for IWMHeaderInfo3 ( hr=0x%08x ).\n" ), hr );
                break;
            }
        }
    }
    while( FALSE );

	return( hr );
}

//------------------------------------------------------------------------------
// Name: CharFilter()
// Desc: Replaces illegal characters "/\*?<>|: with spaces
//------------------------------------------------------------------------------
WCHAR * CharFilter(WCHAR *pwszVar)
{
	size_t pos;
	size_t len = wcsnlen(pwszVar, BUFSIZE);
	do 
	{
		pos = wcscspn( pwszVar,  L"\"/\\*?<>|:" );
		if (pos < len)
			_wcsnset_s( (WCHAR *)(pwszVar+pos),   len,  L' ', 1 );
		else
			break;
	}
	while (pos < len);

	return (pwszVar);
}


//------------------------------------------------------------------------------
// Name: GetAttributes()
// Desc: Retrieves an attribute,
//       using IWMHeaderInfo3.
//------------------------------------------------------------------------------
HRESULT GetAttribute(WCHAR *pwszVarName, 
					 WCHAR *pwszVarValue, 
					 BOOL *hasValue,
					 IWMHeaderInfo * pHeaderInfo, 
					 IWMHeaderInfo3 * pHeaderInfo3)
{
    HRESULT             hr              = S_OK;

    WORD               wAttribNameLen   = 0;
    WORD               wLanguageIndex   = 0;
    DWORD              dwAttribValueLen = 0;
	WORD			   wStreamNum		= 0; 
	WMT_ATTR_DATATYPE  wAttribType		= WMT_TYPE_STRING; 
	WORD cbLength; 

	// 
	// First, retrieve the length of the string, and allocate memory. 
	// 

	hr = pHeaderInfo->GetAttributeByName( &wStreamNum, pwszVarName, &wAttribType, NULL, &cbLength ); 

	if( FAILED( hr ) )
	{ 
		pwszVarValue[0] = L'\0'; 
		return (S_OK);								// no value is not an error
	} 

	//WCHAR * pwszLocalValue = (WCHAR *) new BYTE[ cbLength ]; 
	WCHAR pwszLocalValue[BUFSIZE]; 

	if( !pwszLocalValue ) 
	{ 
		return( E_OUTOFMEMORY ); 
	} 
	//               
	//
	// Now, retrieve the string itself.
	//
	hr = pHeaderInfo->GetAttributeByName( &wStreamNum, pwszVarName, &wAttribType, (BYTE *) pwszLocalValue, &cbLength );

	if( !FAILED( hr ) && (wcsnlen(pwszLocalValue, BUFSIZE) > 0) )
	{
		CharFilter(pwszLocalValue);

		wcscpy_s(pwszVarValue, cbLength, pwszLocalValue );
		*hasValue = true;
	}

	// dealloc
    //delete pwszLocalValue;
	//pwszLocalValue = NULL;

    return( hr );
}


//------------------------------------------------------------------------------
// Name: StandardTree()
// Desc: Puts MP3 files into standard tree
//------------------------------------------------------------------------------
HRESULT StandardTree(WCHAR * wszFilePath, WCHAR * wszAuthor, WCHAR * wszAlbum)
{
    HRESULT hr          = S_OK;

	TCHAR retPath[BUFSIZE];
	WCHAR fromPath[BUFSIZE] = L"..\\..\\";

	do
	{

		hr = GetCurrentDirectory(BUFSIZE, retPath);
        if( FAILED( hr ) )
        {
			_tprintf( _T( "Could not obtain current directory ( hr=0x%08x ).\n" ), hr );
			hr = ERROR_PATH_NOT_FOUND | -1L;
            break;
        }

		// Artist Directory
		if( !SetCurrentDirectoryW(wszAuthor))
		{
			if (! CreateDirectoryW(wszAuthor, NULL) )
			{
				_tprintf( _T( "Could not create Artist directory %ws.\n" ), wszAuthor );
				hr = ERROR_CANNOT_MAKE | -1L;
				break;
			}
			if( !SetCurrentDirectoryW(wszAuthor))
			{
				_tprintf( _T( "Could not change into Artist directory %ws.\n" ), wszAuthor );
				hr = ERROR_CANNOT_MAKE | -1L;
				break;
			}
		}

		// Album Directory
		if( !SetCurrentDirectoryW(wszAlbum) )
		{
			if (! CreateDirectoryW( wszAlbum, NULL) )
			{
				_tprintf( _T( "Could not create Album directory %ws.\n" ), wszAlbum );
				hr = ERROR_CANNOT_MAKE | -1L;
				break;
			}
			if( !SetCurrentDirectoryW( wszAlbum))
			{
				_tprintf( _T( "Could not change into Album directory %ws.\n" ), wszAlbum );
				hr = ERROR_CANNOT_MAKE | -1L;
				break;
			}
		}

		// now move the file

		
		wcscat_s(fromPath, BUFSIZE, wszFilePath);

		if ( !MoveFileW( fromPath, wszFilePath) )
		{
			_tprintf( _T( "Could not move file %ws into Album directory.\n" ), wszFilePath );
			hr = ERROR_CANNOT_MAKE | -1L;
			break;
		}
  

		if ( !SetCurrentDirectory(retPath) )
        {
			_tprintf( _T( "Could not return to original directory.\n" ));
			hr = ERROR_NOT_FOUND | -1L;
            break;
        }

	} while (FALSE);


    return( hr );
}

//------------------------------------------------------------------------------
// Name: Usage()
// Desc: Displays the command-line usage.
//------------------------------------------------------------------------------
void Usage()
{
    _tprintf( _T( "MP3Filer\t [format option] [verbose option]\n" ) );
    _tprintf( _T( "No Copyright (C) Christian Donner. All rights granted.\n" ) );
	_tprintf( _T( "\tFormat options:\n" ) );
	_tprintf( _T( "\t\t -s|-standard: .\\Artist\\Album\\<files>\n" ) );
	_tprintf( _T( "\tFuture enhancements:\n" ) );
	_tprintf( _T( "\t\t -y|-year:     .\\Artist\\Year\\Album\\<files>\n" ) );
	_tprintf( _T( "\t\t -a|-artist:   .\\Artist\\<files>\n" ) );
	_tprintf( _T( "\t\t -b|-album:    .\\Album\\<files>\n" ) );
	_tprintf( _T( "\t\t -g|-genre:    .\\Genre\\Artist\\Album\\<files>\n" ) );
	_tprintf( _T( "\tVerbose options:\n" ) );
	_tprintf( _T( "\t\t -q|-quick:    suppress details\n" ) );
	_tprintf( _T( "\t\t -v|-verbose:  show details\n" ) );
}


//------------------------------------------------------------------------------
// Name: WorkOnFile()
// Desc: Puts MP3 files into standard tree
//------------------------------------------------------------------------------
HRESULT WorkOnFile(int iType, WCHAR * wszFilePath)
{
	HRESULT hr          = S_OK;

	IWMMetadataEditor   * pEditor       = NULL;
    IWMHeaderInfo3      * pHeaderInfo3  = NULL;
    IWMHeaderInfo		* pHeaderInfo   = NULL;

	WCHAR wszAuthor[255];
	WCHAR wszAlbum[255];
	WCHAR wszGenre[255];
	WCHAR wszYear[255];

	BOOL bHasAuthor = false;
	BOOL bHasAlbum = false;
	BOOL bHasGenre = false;
	BOOL bHasYear = false;

    do
    {
        hr = EditorOpenFile( wszFilePath, &pEditor, &pHeaderInfo, &pHeaderInfo3 );
        if( FAILED( hr ) )
        {
            break;
        }

		hr = GetAttribute( L"Author", (WCHAR *)&wszAuthor, &bHasAuthor, pHeaderInfo, pHeaderInfo3);
		if( FAILED( hr ) )
		{
			break;
		}

		hr = GetAttribute( L"WM/AlbumTitle", (WCHAR *)&wszAlbum, &bHasAlbum, pHeaderInfo, pHeaderInfo3);
		if( FAILED( hr ) )
		{
			break;
		}

		hr = GetAttribute( L"WM/Genre", (WCHAR *)&wszGenre, &bHasGenre, pHeaderInfo, pHeaderInfo3);
		if( FAILED( hr ) )
		{
			break;
		}

		hr = GetAttribute( L"WM/Year", (WCHAR *)&wszYear, &bHasYear, pHeaderInfo, pHeaderInfo3);
		if( FAILED( hr ) )
		{
			break;
		}


		if( TYPE_STANDARD == iType )
		{
			//_tprintf( _T("Author: %ws, Album: %ws, Year: %ws, Genre: %ws \n"), &wszAuthor, &wszAlbum, &wszYear, &wszGenre);

			if ( bHasAuthor && bHasAlbum )
			{
				hr = StandardTree( wszFilePath, wszAuthor, wszAlbum );
			}
			else if (!bHasAuthor)
				_tprintf( _T( "Skipped %ws: The file does not have an Author tag.\n" ), wszFilePath);
			else if (!bHasAlbum)
				_tprintf( _T( "Skipped %ws: The file does not have an Album tag.\n" ), wszFilePath);

			if(FAILED( hr ) )
			{
				break;
			}
		}
		else if( TYPE_YEAR == iType )
				_tprintf( _T( "Operation -year not implemented.\n" ) );
		else if( TYPE_ARTIST == iType )
				_tprintf( _T( "Operation -artist not implemented.\n" ) );
		else if( TYPE_ALBUM == iType )
				_tprintf( _T( "Operation -album not implemented.\n" ) );
		else if( TYPE_GENRE == iType )
				_tprintf( _T( "Operation -genre not implemented.\n" ) );
		else
		{
			Usage();
			hr = E_INVALIDARG;
			break;
		}
	}
	while( FALSE );

	( pEditor )->Close();
	/*hr = ( pEditor )->Close();
	if( FAILED( hr ) )
	{
		_tprintf( _T( "Could not close the file %ws ( hr=0x%08x ).\n" ), wszFilePath, hr );
	}*/


    SAFE_RELEASE( pHeaderInfo3 );
    SAFE_RELEASE( pHeaderInfo );
    SAFE_RELEASE( pEditor );

	return hr;
}


//------------------------------------------------------------------------------
// Name: _tmain()
// Desc: Entry point for the application. Processes command-line arguments.
//------------------------------------------------------------------------------
int __cdecl _tmain( int argc, TCHAR * argv[] )
{
    if( argc != 3)
    {
        Usage();
        return( E_INVALIDARG );
    }

	int iType = TYPE_STANDARD;
	int iVerbose = V_QUICK;

	if (      (0 == _tcsicmp( argv[1], _T( "-s" ) )) || 0 == _tcsicmp( argv[1], _T( "-standard" ) )  )  iType = TYPE_STANDARD;
	else if ( (0 == _tcsicmp( argv[1], _T( "-y" ) )) || 0 == _tcsicmp( argv[1], _T( "-year" ) )  )		iType = TYPE_YEAR;
	else if ( (0 == _tcsicmp( argv[1], _T( "-a" ) )) || 0 == _tcsicmp( argv[1], _T( "-artist" ) )  )    iType = TYPE_ARTIST;
	else if ( (0 == _tcsicmp( argv[1], _T( "-b" ) )) || 0 == _tcsicmp( argv[1], _T( "-album" ) )  )     iType = TYPE_ALBUM;
	else if ( (0 == _tcsicmp( argv[1], _T( "-g" ) )) || 0 == _tcsicmp( argv[1], _T( "-genre" ) )  )     iType = TYPE_GENRE;
	else
    {
        Usage();
        return( E_INVALIDARG );
    }

	if (      (0 == _tcsicmp( argv[2], _T( "-v" ) )) || 0 == _tcsicmp( argv[2], _T( "-verbose" ) )  )   iVerbose = V_VERBOSE;
	else if ( (0 == _tcsicmp( argv[2], _T( "-q" ) )) || 0 == _tcsicmp( argv[2], _T( "-quick" ) )  )		iVerbose = V_QUICK;
	else
    {
        Usage();
        return( E_INVALIDARG );
    }


    HRESULT hr              = S_OK;

    struct _wfinddata_t c_file;
    long hFile;

	if( (hFile = _wfindfirst( L"*.mp3", &c_file )) == -1L )
       _tprintf( _T( "No *.mp3 files in current directory!\n" ) );
    else
    {
		do
		{
			if (iVerbose == V_VERBOSE)
				_tprintf( _T( "Processing File: %ws\n" ), c_file.name );

			hr = WorkOnFile(iType, (WCHAR *)c_file.name);
        }
        while( (!FAILED( hr ) ) && (_wfindnext( hFile, &c_file ) == 0) );

       _findclose( hFile );
   }


   if(FAILED( hr ) )
   {
		LPTSTR s;

		
		if ( 0 == FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_ALLOCATE_BUFFER, NULL, hr, 0, (LPTSTR)&s, 0, NULL) )
			_tprintf( _T("Failure - Unknown error\n") );
		else
			_tprintf( _T("%s\n"), s);

		LocalFree(s);
   }

   return( hr );
	}
